import logging
import numpy as np
import os
import pandas as pd
import pickle
import random

import pyarrow.parquet as pq
from descriptors import cachedproperty

from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import GridSearchCV
from sklearn.feature_selection import VarianceThreshold

from sklearn.pipeline import Pipeline, FeatureUnion
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import OneHotEncoder, StandardScaler
from sklearn.impute import SimpleImputer, MissingIndicator
from sklearn.feature_selection import VarianceThreshold

from namematch.data_structures.parameters import Parameters
from namematch.model_evaluation_functions import evaluate_models, get_cv_metrics
from namematch.predict import Predict
from namematch.base import NamematchBase
from namematch.utils.utils import (
    load_parquet_list,
    dump_yaml,
    log_runtime_and_memory,
    determine_model_to_use,
    load_models,
)
from namematch.utils.profiler import Profiler

profile = Profiler()
logger = logging.getLogger()


# globals:
MATCH_COL = 1
SELECTION_OUTCOME = 'match_train_eligible'
MATCH_OUTCOME = 'label'


class FitModel(NamematchBase):
    '''
    Args:
        params (Parameters object): contains parameter values
        all_names_file (str): path to the all-names file
        data_rows_dir (str): path to the data-rows dir
        model_info_file (str): path to the model info yaml file
        output_dir (str): path to the model dir
        traiend_model_info_file (str): path to the model info yaml file of a previously trained model
        selection_model_name (str): selection model name
        match_model_name (str): match model name
        flipped0_file (str): flipped0 file path
    '''
    def __init__(
        self,
        params,
        all_names_file,
        data_rows_dir,
        model_info_file,
        output_dir,
        trained_model_info_file='None',
        selection_model_name='basic_selection_model.pkl',
        match_model_name='basic_match_model.pkl',
        flipped0_file=None,
        *args,
        **kwargs
    ):

        super(FitModel, self).__init__(params, None, *args, **kwargs)

        self.all_names_file = all_names_file
        self.data_rows_dir = data_rows_dir
        self.model_info_file = model_info_file
        self.trained_model_info_file = trained_model_info_file
        self.selection_model_name = selection_model_name
        self.match_model_name = match_model_name
        self.output_dir = output_dir
        self.flipped0_file = flipped0_file if flipped0_file else os.path.join(os.path.dirname(self.output_dir), 'flipped0_potential_edges.csv')
        self.model_info = None
        self.match_models = None
        self.selection_models = None

    @property
    def output_files(self):
        output_files = [self.flipped0_file]
        if self.trained_model_info_file == 'None':
            output_files += [self.model_info_file, os.path.join(self.output_dir, self.match_model_name)]

        if self.params.weight_using_selection_model:
            output_files.append(os.path.join(self.output_dir, self.selection_model_name))

        if self.params.missingness_model is not None:
            if self.params.weight_using_selection_model:
                output_files.append(os.path.join(self.output_dir, f'no_{self.params.missingness_model}_selection_model.pkl'))
            output_files.append(os.path.join(self.output_dir, f'no_{self.params.missingness_model}_match_model.pkl'))

        return output_files

    @property
    def dr_file_list(self):
        return [
            os.path.join(self.data_rows_dir, dr_file) for dr_file in os.listdir(self.data_rows_dir)
        ]

    # @log_runtime_and_memory
    @profile
    def main(self, **kw):
        '''Train and evaluate random foreset model(s). Depending on the settings, this might involved
        training and evaluating multiple types of models (e.g. selection and match models) and/or models
        for different data-row types (e.g. basic and no-dob).
        '''

        if not os.path.exists(self.output_dir):
            os.mkdir(self.output_dir)

        table = pq.read_table(self.all_names_file)
        an = table.to_pandas()
        an = an[an.drop_from_nm == 0]
        an = an.drop_duplicates(['record_id'])

        # add "is valid for training" flag
        an['valid_for_match_train'] = self.find_valid_training_records(an,
                self.params.match_train_criteria.get('all_names', {}))
        an_train_eligible_dict = an.set_index('record_id')['valid_for_match_train'].to_dict()

        # remove flipped0 file if it exists (since this only sometimes gets created,
        # don't want it lingering for future runs -- messes up potential_edges)
        if os.path.exists(self.flipped0_file):
            os.remove(self.flipped0_file)

        # if a trained model is provided, load it
        if self.trained_model_info_file != "None":
            self.match_models, self.model_info = load_and_save_trained_model(
                self.trained_model_info_file,
                self.model_info_file
            )
            match_train_df, match_eval_df, prob_match_train = self.get_train_eval_data(
                None,
                self.model_info,
                self.params,
                'match',
                any_train=False
            )

            match_eval_phats = Predict.predict(self.match_models, match_eval_df, 'match')
            flipped0 = get_flipped0_potential_edges(
                match_eval_phats,
                self.model_info,
                self.params.allow_clusters_w_multiple_unique_ids
            )
            flipped0.to_csv(self.flipped0_file, index=False)

            return

        # if a trained model is not provided, need to build it
        else:
            self.model_info = define_necessary_models(
                self.dr_file_list,
                self.output_dir,
                self.params.missingness_model,
                selection_model_name=self.selection_model_name,
                match_model_name=self.match_model_name,
            )
            if self.params.weight_using_selection_model:
                selection_train_df, selection_eval_df, prob_match_train = self.get_train_eval_data(
                    None, self.model_info, self.params, 'selection')

                self.selection_models = self.fit_models(
                        selection_train_df, 'selection', self.model_info)

                selection_eval_phats = Predict.predict(self.selection_models, selection_eval_df,
                        'selection', prob_match_train=prob_match_train)
                self.evaluate_models(selection_eval_phats, SELECTION_OUTCOME, 'selection')

                selection_train_df.to_parquet(os.path.join(self.output_dir, 'selection_train_df.parquet')) # TEMP
                selection_eval_df.to_parquet(os.path.join(self.output_dir, 'selection_eval_df.parquet')) # TEMP

            match_train_df, match_eval_df, prob_match_train = self.get_train_eval_data(
                an_train_eligible_dict,
                self.model_info,
                self.params,
                'match'
            )

            if self.params.weight_using_selection_model:
                # needed for weighting
                # NOTE: some of the weights will be in-sample
                match_train_df = Predict.predict(self.selection_models, match_train_df,
                        'selection', all_cols=True, all_models=True, prob_match_train=prob_match_train)
                match_eval_df = Predict.predict(self.selection_models, match_eval_df,
                        'selection', all_cols=True, all_models=True, prob_match_train=prob_match_train)

            self.match_models = self.fit_models(
                    match_train_df, 'match', self.model_info)

            match_train_phats = Predict.predict(self.match_models, match_train_df, 'match', oob=True)
            match_eval_phats = Predict.predict(self.match_models, match_eval_df, 'match')
            match_train_phats.to_parquet(os.path.join(self.output_dir, 'train_phats.parquet')) # TEMP
            match_eval_phats.to_parquet(os.path.join(self.output_dir, 'eval_phats.parquet')) # TEMP

            match_train_df.filter(regex='record_id|weight').to_parquet(os.path.join(self.output_dir, 'match_train_weights.parquet')) # TEMP
            match_eval_df.filter(regex='record_id|weight').to_parquet(os.path.join(self.output_dir, 'match_eval_weights.parquet')) # TEMP

            thresholds_dict = self.evaluate_models(
                match_eval_phats,
                MATCH_OUTCOME, 'match',
                self.params.weight_using_selection_model,
                self.params.default_threshold,
                self.params.missingness_model_threshold_boost,
                self.params.optimize_threshold,
                self.params.fscore_beta,
            )

            self.model_info = add_threshold_dict(self.model_info, thresholds_dict)

            flipped0 = get_flipped0_potential_edges(
                    pd.concat([match_train_phats, match_eval_phats]),
                    self.model_info,
                    self.params.allow_clusters_w_multiple_unique_ids)

            # save models and info
            dump_yaml(self.model_info, self.model_info_file)
            save_models(self.selection_models, self.match_models, self.model_info)

            flipped0.to_csv(self.flipped0_file, index=False)

            if self.enable_lprof:
                self.write_line_profile_stats(profile.line_profiler)


    # @log_runtime_and_memory
    def fit_model(self, df, vars_to_exclude, outcome, weights=None, n_jobs=1, **kw):
        '''Fit random forest model.

        Args:
            df (pd.DataFrame): data rows, subset to training rows
            vars_to_exclude (list): variables to disallow from the model
            outcome (string): name of the column that we're predicting
            weights (list): sample weights to use for training (can be None)
            n_jobs (int): number of jobs to run in parallel

        Returns:
                tuple: tuple containing:

                - **mod** (*sklearn.ensemble.RandomForestClassifier*): trained sklearn random forest model object
                - **feature_info** (*pd.DataFrame*): feature_importance
        '''

        if df[outcome].nunique() == 1:
            logger.error('Something went wrong; all training outcomes are {}'.format(df[outcome].tolist()[0]))
            raise ValueError

        col_names = df.columns.values.tolist()
        col_names = [col for col in col_names if col not in vars_to_exclude]
        numeric_cols = [c for c in col_names if 'var_' in c and '_missing' not in c and 'match' not in c]
        categorical_cols = [c for c in col_names if 'var_' in c  and ('_missing' in c or 'match' in c)]

        numeric_transformer = FeatureUnion([
            ('impute_numeric', Pipeline(steps=[
                ('imputer', SimpleImputer(strategy='mean'))])),
            ('missing_dummies', Pipeline(steps=[
                ('missing', MissingIndicator(error_on_new=False))]))
        ])

        categorical_transformer = Pipeline(steps=[
            ('imputer', SimpleImputer(strategy='constant', fill_value=-1)),
            ('onehot', OneHotEncoder(handle_unknown='ignore'))])

        preprocessor = ColumnTransformer(
            transformers=[
                ('num', numeric_transformer, numeric_cols),
                ('cat', categorical_transformer, categorical_cols)],
            sparse_threshold=0)

        base_clf = RandomForestClassifier(
                n_estimators=200,
                oob_score=True,
                n_jobs=n_jobs,
                verbose=0)

        pipeline = Pipeline([
            ('preprocessor', preprocessor),
            ('no_variance_filter', VarianceThreshold()),
            ('clf', base_clf)
        ])

        pipeline_params = {
            'clf__min_samples_leaf':[25, 150],
        }

        mod = GridSearchCV(pipeline, pipeline_params,
			cv=3,
			n_jobs=1, verbose=0,
			scoring='f1',
            return_train_score=True
        )

        mod.fit(df, df[outcome],
                clf__sample_weight=weights)

        feature_info = get_feature_info(mod.best_estimator_, numeric_cols, categorical_cols)

        return mod, feature_info

    def fit_models(self, train_df, model_type, model_info):
        '''Fit random forest model.

        Args:
            train_df (pd.DataFrame): data rows, subset to training rows
            model_type (string): either "selection" or "match"
            model_info (dict): dict with information about how to fit the model

        Returns:
            dict: maps model name (e.g. basic or no_dob) to a trained model object
        '''

        models = {}
        feature_infos = {}
        cv_metrics = {}
        for this_model_name, this_model_info in model_info.items():

            outcome = MATCH_OUTCOME
            if model_type == 'selection':
                outcome = SELECTION_OUTCOME

            weights = None
            if model_type == 'match':
                if f'{this_model_name}_weight' in train_df.columns:
                    weights = train_df[f'{this_model_name}_weight']
                    logger.info(f'Using {this_model_name} weights.')

            logger.info(f'Training {this_model_name} model ({model_type}).')

            models[this_model_name], feature_infos[this_model_name] = self.fit_model(
                train_df,
                this_model_info['vars_to_exclude'],
                outcome,
                weights=weights,
                n_jobs=self.params.num_workers)

        logger.stat_dict({f"feature_info__{model_type}": feature_infos})
        self.stats_dict[f'feature_info__{model_type}'] = feature_infos

        return models

    def evaluate_models(self,
            phats_df,
            outcome,
            model_type,
            weight_using_selection_model=False,
            default_threshold=0.5,
            missingness_model_threshold_boost=0.2,
            optimize_threshold=False,
            fscore_beta=1.0,
            **kw
            ):
        return evaluate_models(
                    phats_df,
                    outcome,
                    model_type,
                    weight_using_selection_model,
                    default_threshold,
                    missingness_model_threshold_boost,
                    optimize_threshold,
                    fscore_beta,
                    self.stats_dict,
                    )

    def get_train_eval_data(self, an_train_eligible_dict, model_info, params, model_type, any_train=True):
        '''Load data-rows, filter to rows that are eligible for training a givem model type, and then split
        the data into a training set and a labeled evaulation set.

        Args:
            an_train_eligible_dict (dict): maps record_id to flag indicating record's all-names based training eligibility
            model_info (dict): dict with information about how to fit the model
            params (Parameters object): contains parameter values
            model_type (str): either "selection" or "match"
            any_train (bool): True if you want training data (e.g. not a pre-trained model), False otherwise

        Returns:
            pd.DataFrame: data rows, filtered to training data (excluding labeled eval data)
            pd.DataFrame: data rows, filtered to labeled eval data
            float: share of data rows that are labeled
        '''

        temp_dr = load_parquet_list(
            self.dr_file_list,
            cols=['dr_id', 'covered_pair', 'labeled_data'])
        n_data_rows = len(temp_dr)
        logger.info(f"Number of data rows: {n_data_rows}")
        self.stats_dict['n_data_rows'] = n_data_rows
        n_covered_data_rows = len(temp_dr[temp_dr.covered_pair == 1])
        self.stats_dict['n_covered_data_rows'] = n_covered_data_rows
        del temp_dr

        if model_type == 'selection':
            sample = params.max_selection_train_eval_n / n_data_rows
            if sample > 1:
                sample = 1
            conditions_dict = {'covered_pair': 1}
            outcome = SELECTION_OUTCOME
        elif model_type == 'match':
            sample = 1
            conditions_dict = {'labeled_data':1, 'covered_pair': 1}
            outcome = MATCH_OUTCOME
        else:
            logger.error("Invalid model_type supplied: must be either 'selection' or 'match.'")
            raise

        train_eval_df = load_parquet_list(
                self.dr_file_list,
                conditions_dict=conditions_dict,
                sample=sample)

        # create match_train_eligible flag (1 if labeled and meet data-row and all-names critiera, 0 otherwise)
        dr_train_eligible_conditions_dict = params.match_train_criteria.get('data_rows', {})
        train_eval_df['match_train_eligible'] = \
            get_match_train_eligible_flag(train_eval_df, dr_train_eligible_conditions_dict, an_train_eligible_dict)

        if model_type == 'selection':
            prob_match_train = (train_eval_df.match_train_eligible.sum() / sample) / self.stats_dict['n_covered_data_rows']
            self.stats_dict['prob_match_train'] = float(prob_match_train)
            logger.info(f"P(s): {prob_match_train}")
        else:
            prob_match_train = self.stats_dict.get('prob_match_train', None)

        train_eval_df[outcome] = train_eval_df[outcome].astype(int)

        train_eval_df['model_to_use'] = \
                determine_model_to_use(train_eval_df, model_info, verbose=True)

        if not any_train:
            train_df = None
            eval_df = train_eval_df.copy()

        else: # need train/eval split

            pct_train = params.pct_train
            if (model_type == 'match') and (pct_train * len(train_eval_df) > params.max_match_train_n):
                pct_train = params.max_match_train_n / len(train_eval_df)

            train_eval_df['train'] = [random.random() <= pct_train
                                      for r in range(len(train_eval_df))]
            # TODO should we split at uid level? -- hard since these are pairs?

            if model_type == 'selection':
                train_df = train_eval_df[(train_eval_df.train == 1)].copy()
                eval_df = train_eval_df[(train_eval_df.train == 0)].copy()
            elif model_type == 'match':
                train_df = train_eval_df[
                        (train_eval_df.train == 1) &
                        (train_eval_df.match_train_eligible == 1)].copy()
                eval_df = train_eval_df[
                        (train_eval_df.train == 0) &
                        (train_eval_df.match_train_eligible == 1)].copy()

        return train_df, eval_df, prob_match_train


    # @log_runtime_and_memory
    def find_valid_training_records(self, an, an_match_criteria):
        '''Identify records that meet the all-names criteria for training data.

        Args:
            an (pd.DataFrame): all-names table (one row per input record)

                =====================   =======================================================
                record_id               unique record identifier
                file_type               either "new" or "existing"
                <fields for matching>   both for the matching model and for constraint checking
                <raw name fields>       pre-cleaning version of first and last name
                blockstring             concatenated version of blocking columns (sep by ::)
                drop_from_nm            flag, 1 if met any "to drop" criteria 0 otherwise
                =====================   =======================================================

            an_match_criteria (dict): keys are all-names columns, mapped to acceptable values

        Returns:
            pd.Series: flag, 1 if the record is eligible for training set 0 otherwise
        '''

        an['meets_match_criteria'] = 1
        for col, accepted_values in an_match_criteria.items():
            if not isinstance(accepted_values, list):
                accepted_values = [accepted_values]
            accepted_values = [str(av) for av in accepted_values]
            an['meets_match_criteria'] = (an['meets_match_criteria']) & (an[col].isin(accepted_values))

        return an['meets_match_criteria']


def get_feature_info(pipeline, raw_num_cols, raw_cat_cols):
    '''Extract the feature importance information from a sklearn model pipeline.

    Args:
        pipeline (skleran fitted pipeline): trained model
        raw_num_cols (list): numeric columns that went into the model (before pipeline processing)
        raw_cat_cols (list): categorical columns that went into the model (before pipeline processing)

    Returns:
        pd.DataFrame: feature importance information

        ==============   =================================================
        feature          name of the feautre
        importance       relative importance of this feature to the model
        ==============   =================================================
    '''

    raw_num_cols = raw_num_cols[:]
    raw_cat_cols = raw_cat_cols[:]

    preprocessor = pipeline.named_steps['preprocessor']
    num_transf = preprocessor.named_transformers_['num']
    cat_transf = preprocessor.named_transformers_['cat']

    num_cols_imputed = [raw_num_cols[col_ix]
                        for col_ix in range(len(raw_num_cols))
                        if not np.isnan(num_transf.transformer_list[0][1].named_steps['imputer'].statistics_[col_ix])]

    num_cols_missing = [raw_num_cols[col_ix] + '_is_missing'
                        for col_ix in num_transf.transformer_list[1][1].named_steps['missing'].features_]

    cat_feature_names = ['%s_%s' % (raw_cat_cols[col_ix], cat)
                         for col_ix in range(len(raw_cat_cols))
                         for cat in cat_transf.named_steps['onehot'].categories_[col_ix]]

    feature_names_pre_varthresh = num_cols_imputed + num_cols_missing + cat_feature_names

    retained_flags = pipeline.named_steps['no_variance_filter'].get_support()
    feature_names = [feature_names_pre_varthresh[i]
                     for i in range(len(feature_names_pre_varthresh)) if retained_flags[i]]
    dropped_features = [feature_names_pre_varthresh[i]
                        for i in range(len(feature_names_pre_varthresh)) if not retained_flags[i]]

    importances = [float(imp) for imp in list(pipeline.named_steps['clf'].feature_importances_)]

    feature_info = {"retained": feature_names, "imp": importances, "dropped": dropped_features}

    return feature_info


def save_models(selection_models, match_models, model_info):
    '''Save the models to file.

    Args:
        selection_models (dict): maps model name (e.g. basic or no-dob) to a fit match model object
        match_models (dict): maps model name (e.g. basic or no-dob) to a fit selection model object
        model_info (dict): dict with information about how to fit the model
    '''

    for model_name, this_model_info in model_info.items():

        # save selection model
        if selection_models is not None:
            with open(this_model_info['selection_model_path'], 'wb') as mf:
                pickle.dump(selection_models[model_name], mf)

        # save match model
        with open(this_model_info['match_model_path'], 'wb') as mf:
            pickle.dump(match_models[model_name], mf)


def define_necessary_models(
        dr_file_list,
        output_dir,
        missing_mod_field=None,
        selection_model_name='basic_selection_model.pkl',
        match_model_name='basic_match_model.pkl',
        ):
    '''Determine the different models needed (using a sample) and
    define the characteristics of data that determine which
    model should handle it.

    NOTE: Right now, there is an assumption that the training universe
          is the same between all models (i.e. basic and missingness)

    Args:
        dr_file_list (list): list of paths to all data row files
        output_dir (str): model output folder path
        missing_mod_field (str or None): field that could trigger need for separate model

    Returns:
        dict: mapping the name of a model (str) to a dict of the following information:

            - selection_model_name (str)
            - match_model_name (str)
            - type (str): one of "default" or "missingness"
            - actual_phat_universe (dict): maps a variable name to a value(?)
            - vars_to_exclude (str list)
            - match_thresh (float): threshold for match/nonmatch
    '''

    # initialize categories (always need a basic model)
    categories = {
        'basic' : {
            'selection_model_path' :
                os.path.abspath(os.path.join(output_dir, selection_model_name)),
            'match_model_path' :
                os.path.abspath(os.path.join(output_dir, match_model_name)),
            'type' : 'default',
            'actual_phat_universe' : {},
            'vars_to_exclude' : [],
            'match_thresh' : None
        }
    }

    # determine if missingness model is needed
    if missing_mod_field is not None:
        # missingness model is needed; universe is same as "basic" model

        var_regex = f'var_{missing_mod_field}_'
        dr_first_row = pd.read_parquet(dr_file_list[0], engine='pyarrow').head(1)
        related_variables = dr_first_row.filter(regex=var_regex).columns.tolist()

        categories[f'no_{missing_mod_field}'] = {
            'selection_model_path' :
                os.path.abspath(os.path.join(output_dir, f'no_{missing_mod_field}_selection_model.pkl')),
            'match_model_path' :
                os.path.abspath(os.path.join(output_dir, f'no_{missing_mod_field}_match_model.pkl')),
            'type' : 'missingness',
            'actual_phat_universe' : {
                f'var_{missing_mod_field}_missing' : 1
            },
            'vars_to_exclude' : related_variables,
            'match_thresh' : None
        }

    # NOTE: will need more logic here once we have model types other than
    #       basic and missing (e.g. uncovered_pairs)

    return categories


def load_and_save_trained_model(trained_model_info_file, output_file):
    '''Load a set of pre-trained models and copy them to the current run's
    output directory. Typically only used in incremental runs.

    Args:
        trained_model_info_file (str): path to a model yaml file, which has path/threshold/universe info
        output_file (str): path to output the current run's model yaml file (for copying)

    Returns:
        dict: maps model name (e.g. basic or no_dob) to a trained model object
    '''

    try:
        selection_models, \
        match_models, \
        model_info = load_models(trained_model_info_file, selection=True)
        if len(match_models) == 0:
            raise
    except:
        logger.error("Unable to load trained model or its info yaml.")
        raise

    dump_yaml(model_info, output_file)

    # NOTE: not returning selection_models now, but may need to once using
    #       pre-trained model is fully functional (for unbiased evaluation)
    return match_models, model_info


def get_match_train_eligible_flag(df, dr_train_eligible_conditions_dict, an_train_eligible_dict):
    '''Determine if a data-row is eligible for training (for match models), according
    to both all-names eligibility criteria and data-row eligibility criteria.

    Args:
        df (pd.DataFrame): portion of data-rows file, limited to labeled rows
        dr_train_eligible_conditions_dict (dict): contains data-row training eligibility criteria
        an_train_eligible_dict (dict): maps record_id to flag indicating record's all-names based training eligibility

    Returns:
        pd.Series: flag, 1 if data-row is training eligible (for match models)
    '''

    df = df.copy()

    df['match_train_eligible'] = df.labeled_data

    if an_train_eligible_dict is not None:

        df.loc[df.match_train_eligible == 1, 'record_1_elig'] = \
                df[df.match_train_eligible == 1].record_id_1.map(an_train_eligible_dict)
        df.loc[df.match_train_eligible == 1, 'record_2_elig'] = \
                df[df.match_train_eligible == 1].record_id_2.map(an_train_eligible_dict)

        df['match_train_eligible'] = \
                ((df.match_train_eligible) &
                 (df.record_1_elig == 1) &
                 (df.record_2_elig == 1)).astype(int)

    for col, accepted_values in dr_train_eligible_conditions_dict.items():

        if not isinstance(accepted_values, list):
            accepted_values = [accepted_values]

        df['match_train_eligible'] = \
                (df['match_train_eligible'] &
                    df[col].isin(accepted_values))

    return df.match_train_eligible


def add_threshold_dict(model_info, thresholds_dict):
    '''Add threshold information to the model_info dict, once it's been determined.

    Args:
        model_info (dict): dict with information about how to fit the model
        thresholds_dict (dict): keys are model name (e.g. basic, no-dob), values are optimized thresholds

    Return:
        dict: model dict, now with threshold info
    '''

    for model_name, threshold in thresholds_dict.items():
        model_info[model_name]['match_thresh'] = threshold

    return model_info


def get_flipped0_potential_edges(phats_df, model_info, allow_clusters_w_multiple_unique_ids):
    '''If allowed, identify the set of labeled 0s with high phats so they can be treated
    as matches downstream.

    Args:
        phats_df (pd.DataFrame): phat info for record pairs

            =====================   =======================================================
            record_id (_1, _2)      unique record identifiers
            model_to_use            based on pair characteristics, which model to use (e.g. basic or no-dob)
            covered_pair            did the pair make it through blocking
            match_train_eligible    is the pair eligible for training (for match model)
            exactmatch              is the pair an exact match on name/dob
            label                   whether the pair is a match or not
            <phat_col>              predicted probability of match
            =====================   =======================================================

        model_info (dict): dict with information about how to fit the model
        allow_clusters_w_multiple_unique_ids (bool): param controlling if 0s can be flipped to 1

    Returns:
        pd.DataFrame: same as phats_df, just
    '''

    if not allow_clusters_w_multiple_unique_ids:
        return pd.DataFrame(columns=phats_df.columns)

    phats_df['flipped0'] = 0

    for model_name, this_model_info in model_info.items():

        phat_col = f'{model_name}_match_phat'
        phats_df.loc[(phats_df.model_to_use == model_name) &
                     (phats_df.label == 0) &
                     (phats_df[phat_col] >= this_model_info['match_thresh']), 'flipped0'] = 1

    return phats_df[phats_df.flipped0 == 1].drop(columns=['flipped0'])



import editdistance
import importlib.util
import logging
import networkx as nx
import numpy as np
import pandas as pd
import pickle
import os
import time
import gc

from typing import Union
from collections import deque

import pyarrow as pa
import pyarrow.parquet as pq

from namematch.base import NamematchBase
from namematch.data_structures.parameters import Parameters
from namematch.data_structures.schema import Schema
from namematch.utils.utils import log_runtime_and_memory, load_parquet
from namematch.default_constraints import get_columns_used, is_valid_link, is_valid_cluster, apply_link_priority
from namematch.utils.profiler import Profiler


profile = Profiler()

logger = logging.getLogger()

class Constraints(object):

    @property
    def get_columns_used(self):
        try:
            return self._get_columns_used

        except AttributeError:
            logging.warning(f"Constraints.get_columns_used wasn't set. Use default function from {get_columns_used.__module__}")
            return get_columns_used

    @get_columns_used.setter
    def get_columns_used(self, func: callable):
        self._get_columns_used = func

    @property
    def is_valid_link(self):
        try:
            return self._is_valid_link

        except AttributeError:
            logging.warning(f"Constraints.is_valid_link wasn't set. Use default function from {is_valid_link.__module__}")
            return is_valid_link

    @is_valid_link.setter
    def is_valid_link(self, func: callable):
        self._is_valid_link = func

    @property
    def is_valid_cluster(self):
        try:
            return self._is_valid_cluster

        except AttributeError:
            logging.warning(f"Constraints.is_valid_cluster wasn't set. Use default function from {is_valid_cluster.__module__}")
            return is_valid_cluster

    @is_valid_cluster.setter
    def is_valid_cluster(self, func: callable):
        self._is_valid_cluster = func

    @property
    def apply_link_priority(self):
        try:
            return self._apply_link_priority

        except AttributeError:
            logging.warning(f"Constraints.apply_link_priority wasn't set. Use default function from  {apply_link_priority.__module__}")
            return apply_link_priority

    @apply_link_priority.setter
    def apply_link_priority(self, func: callable):
        self._apply_link_priority = func


class Cluster(NamematchBase):
    '''
    Args:
        params (Parameters object): contains parameter values
        schema (Schema object): contains match schema info (files to match, variables to use, etc.)
        constraints (str or Constrants object): either a path to python script defining constraint functions or a Constraints object
        must_links_file (str): path to the must-links file
        potential_edges_dir (str): path to the potential-links dir in the output/details folder
        flipped0_edges_file (str): path to the flipped-links file
        all_names_file (str): path to the all-names file
        cluster_assignments(str): path to the cluster-assignments file
    '''
    def __init__(
        self,
        params,
        schema,
        must_links_file="must_links.csv",
        potential_edges_dir="potential_links",
        flipped0_edges_file="flipped0_potential_links.csv",
        all_names_file="all_names.parquet",
        cluster_assignments="cluster_assignments.pkl",
        edges_to_cluster="edges_to_cluster.parquet",
        constraints: Union[str, Constraints]=None,
        *args,
        **kwargs
    ):
        super(Cluster, self).__init__(params, schema, *args, **kwargs)

        self.constraints = constraints
        self.must_links_file = must_links_file
        self.potential_edges_dir = potential_edges_dir
        self.flipped0_edges_file = flipped0_edges_file
        self.all_names_file = all_names_file
        self.cluster_assignments = cluster_assignments
        self.edges = edges_to_cluster

    @property
    def output_files(self):
        return [self.cluster_assignments, self.edges]

    # @log_runtime_and_memory
    @profile
    def main(self, **kw):
        '''Read the record pairs with high probability of matching and connect them in a way
        that doesn't violate any logic constraints to form clusters.
        '''
        # loading cluster_logic functions
        cluster_logic = self.get_cluster_logic(self.constraints)

        # loading must links
        logger.info("Loading must-link links.")
        must_links_df = pd.read_csv(self.must_links_file, low_memory=False)
        # get uid and eid cols
        uid_cols = self.schema.variables.get_variables_where(
                attr='compare_type', attr_value='UniqueID')
        try:
            eid_col = self.schema.variables.get_variables_where(
                    attr='compare_type', attr_value='ExistingID')[0]
        except:
            eid_col = None

        # load info needed for constraint checking
        logger.info("Creating dictionary of cluster information.")
        cluster_info = self.load_cluster_info(self.all_names_file, uid_cols, eid_col, cluster_logic)

        # separate must-links if we can't initialize new 1s
        if not self.params.initialize_from_ground_truth_1s or self.params.incremental:
            gt_1s_df = must_links_df[must_links_df.existing == 0].copy()
            must_links_df = must_links_df[must_links_df.existing == 1].copy()
        else:
            gt_1s_df = None

        # create a starting point using must-links
        logger.info("Initializing initial clusters.")
        clusters, cluster_assignments, original_cluster_ids = \
                self.get_initial_clusters(must_links_df, cluster_info, eid_col)
        # potential_edges is sorted (decreasing) by phat
        logger.info("Loading potential links.")
        potential_edges_files = [os.path.join(self.potential_edges_dir, pe_file)
                                 for pe_file in os.listdir(self.potential_edges_dir)]
        self.get_potential_edges(
                potential_edges_files, self.flipped0_edges_file, gt_1s_df, cluster_logic,
                cluster_info, uid_cols, eid_col)

        logger.info("Clustering potential links.")
        cluster_assignments = self.cluster_potential_edges(
                clusters, cluster_assignments, original_cluster_ids, cluster_info,
                cluster_logic, uid_cols, eid_col)

        with open(self.cluster_assignments, "wb") as f:
            pickle.dump(cluster_assignments, f)

        if self.enable_lprof:
            self.write_line_profile_stats(profile.line_profiler)

    def get_cluster_logic(self, constraints):
        if isinstance(constraints, str):
            spec = importlib.util.spec_from_file_location("module.name", constraints)
            constraints = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(constraints)
            logger.info(f"Set up cluster constraints {constraints}")

        elif isinstance(constraints, Constraints):
            logger.info(f"Set up cluster constraints {constraints}")
            constraints = self.constraints

        else:
            if constraints is not None:
                logger.warning(f"'constraints' {constraints} is not recognized! Using the default link and cluster constraints.")
            else:
                logger.warning(f"No 'constraints' provided: using the default link constraints, cluster constraints, and edge priority.")
            import namematch.default_constraints as constraints

        cluster_logic = Constraints()
        cluster_logic.get_columns_used = constraints.get_columns_used
        cluster_logic.is_valid_link = constraints.is_valid_link
        cluster_logic.is_valid_cluster = constraints.is_valid_cluster
        cluster_logic.apply_link_priority = constraints.apply_link_priority
        cluster_logic.enable_lprof = self.enable_lprof

        return cluster_logic

    @profile
    def auto_is_valid_edge(
                self,
                edges_df,
                uid_cols, allow_clusters_w_multiple_unique_ids, leven_thresh,
                eid_col=None):
        '''Check if two records would violate a unique id or existing id constraint.

        Args:
            edges_df (pd.DataFrame): potential edges information

                ================      =======================================================
                record_id_1           unique record identifier (for first in pair)
                record_id_2           unique record identifier (for second in pair)
                phat                  predicted probability of a record pair being a match
                original_order        original ordering 1-N (useful so gt is always on top of phat=1 cases)
                ================      =======================================================

            uid_cols (list): all-names column(s) with compare_type UniqueID
            allow_clusters_w_multiple_unique_ids (bool): True if a cluster can have multiple uid values
            leven_thresh (int): n character edits to allow between uids before they're considered different
            eid_col (str): all-names column with compare_type ExistingID (None for non-incremental runs)

        Returns:
            valid_edges_df: potential edges information, but limited to rows that pass the automated validity check
        '''

        edges_df['valid'] = True

        if eid_col is not None:
            # if both existing ids are known, they cannot be different
            both_eids_known_ix = (edges_df[f"{eid_col}_1"].notnull()) & (edges_df[f"{eid_col}_2"].notnull())
            edges_df.loc[both_eids_known_ix & (edges_df[f"{eid_col}_1"] != edges_df[f"{eid_col}_2"]), 'valid'] = False

        if allow_clusters_w_multiple_unique_ids:
            return edges_df[edges_df.valid == True]

        # NOTE if you've made it here, allow_clusters_w_multiple_unique_ids is False
        # if both unique ids are known, they cannot be different
        edges_df['violations'] = 0
        edges_df['attempts'] = 0
        for uid_col in uid_cols:

            both_uids_known_ix = (edges_df[f"{uid_col}_1"].notnull()) & (edges_df[f"{uid_col}_2"].notnull())
            edges_df.loc[both_uids_known_ix, 'attempts'] = edges_df[both_uids_known_ix].attempts + 1

            uids_different_ix = edges_df[f"{uid_col}_1"] != edges_df[f"{uid_col}_2"]

            if leven_thresh is None:
                edges_df.loc[both_uids_known_ix & uids_different_ix, 'violations'] = \
                    edges_df[both_uids_known_ix & uids_different_ix].violations + 1
            else:
                edges_df.loc[both_uids_known_ix & uids_different_ix, f'{uid_col}_edit_dist'] = \
                    np.vectorize(editdistance.eval, otypes=['float'])(
                        edges_df[both_uids_known_ix & uids_different_ix][f"{uid_col}_1"].values,
                        edges_df[both_uids_known_ix & uids_different_ix][f"{uid_col}_2"].values)
                leven_violated = edges_df[f'{uid_col}_edit_dist'] > leven_thresh
                edges_df.loc[both_uids_known_ix & uids_different_ix & leven_violated, 'violations'] = \
                    edges_df[both_uids_known_ix & uids_different_ix & leven_violated].violations + 1

        edges_df.loc[(edges_df.attempts > 0) & (edges_df.attempts == edges_df.violations), 'valid'] = False

        return edges_df[edges_df.valid == True]

    @profile
    def auto_is_valid_cluster(
                self,
                cluster,
                uid_cols, allow_clusters_w_multiple_unique_ids, leven_thresh,
                eid_col=None):
        '''Check if a proposed cluster would violate a unique id or existing id constraint.

        Args:
            cluster (pd.DataFrame): all-names file (relevant columns only) records for the proposed cluster
            uid_cols (list): all-names column(s) with compare_type UniqueID
            allow_clusters_w_multiple_unique_ids (bool): True if a cluster can have multiple uid values
            leven_thresh (int): n character edits to allow between uids before they're considered different
            eid_col (str): all-names column with compare_type ExistingID (None for non-incremental runs)

        Returns:
            bool: False if an automated constraint is violated
        '''

        if eid_col is not None:

            # at most one existing id in a cluster
            n_existing_ids = cluster[eid_col].nunique()
            if n_existing_ids > 1:
                return False

        if allow_clusters_w_multiple_unique_ids:
            return True

        # NOTE if you've made it here, allow_clusters_w_multiple_unique_ids is False

        # at most one unique id in a cluster
        for uid_col in uid_cols:

            n_unique_ids = cluster[uid_col].nunique()
            if n_unique_ids > 1:

                if leven_thresh is None:
                    return False
                else:
                    # make sure that each non-NA uid has one other non-NA uid
                    # within leven_thresh edits
                    cluster.reset_index(inplace=True)
                    uid_df = cluster[~cluster[uid_col].isnull()][[uid_col]]
                    uid_df['dummy'] = 1
                    uid_df = pd.merge(uid_df, uid_df, on='dummy', suffixes = ['_1', '_2'])
                    uid_df = uid_df[uid_df[f'{uid_col}_1'] != uid_df[f'{uid_col}_2']].copy()
                    uid_df['ed'] = np.vectorize(editdistance.eval)(
                            uid_df[f'{uid_col}_1'].values, uid_df[f'{uid_col}_2'].values)
                    min_ed = uid_df.groupby(f'{uid_col}_1').ed.min()
                    return (min_ed > leven_thresh).sum() == 0

        return True

    # @log_runtime_and_memory
    @profile
    def get_initial_clusters(self, must_links_df, an_df, eid_col, **kw):
        '''Use must links (ground truth and/or a previous run) to create the
        starting clusters.

        Args:
            must_links_df (pd.DataFrame): record pairs that must be linked together no matter what

                ==============  ======================================================================
                record_id_1     unique identifier for the first record in the pair
                record_id_2     unique identifier for the second record in the pair
                blockstring_1   blockstring for the first record in the pair
                blockstring_2   blockstring for the second record in the pair
                drop_from_nm_1  flag, 1 if the first record in the pair was not eligible for matching
                drop_from_nm_2  flag, 1 if the second record in the pair was not eligible for matching
                existing        flag, 1 if the pair is must-link because of ExistingID
                ==============  ======================================================================

            an_df (pd.DataFrame): all-names file, with only the columns relevant for clustering

                ===========================  =======================================================
                record_id                    unique record identifier
                <uid column(s)>              columns with compare_type UniqueID
                <eid column(s)>              columns with compare_type ExistingID
                <user-constraint column(s)>  (optional) columns mentioned in `get_columns_used()`
                ===========================  =======================================================

            eid_col (str): all-names column with compare_type ExistingID, or None

        Returns:
            dict: clusters maps a cluster id to a list of record ids
            dict: cluster_assignments maps a record_id to a cluster_id
            set: cluster ids that are already in use (only for incremental)
        '''

        # clusters maps a cluster id to a list of record ids
        # cluster_assignments maps a record_id to a cluster_id
        if eid_col is not None: # incremental

            # make dictionary mapping cluster id to a list of records in that cluster
            eid_df = an_df[an_df[eid_col].notnull() & (an_df[eid_col] != '')].copy()
            eid_df[eid_col] = eid_df[eid_col].astype(int)
            cluster_assignments = eid_df[[eid_col]].to_dict()[eid_col]
            clusters = {}
            for (record_id, cluster_id) in cluster_assignments.items():
                if (cluster_id in clusters):
                    clusters[cluster_id].append(record_id)
                else:
                    clusters[cluster_id] = [record_id]

            original_cluster_ids = set(clusters.keys())

            if len(clusters) > 0:
                # for later assign the cluster_id
                i = max(clusters.keys())

        else: # from scratch
            clusters = {}

            if len(must_links_df) > 0:
                reversed_must_links_df = pd.DataFrame({
                    "record_id_1": must_links_df["record_id_2"],
                    "record_id_2": must_links_df["record_id_1"]
                })
                must_links_df = pd.concat([must_links_df, reversed_must_links_df], join='inner')

                # find the clusters, ie connected components
                g = nx.from_pandas_edgelist(
                        must_links_df,
                        source="record_id_1",
                        target="record_id_2")
                for i, comp in enumerate(nx.connected_components(g)):
                    clusters[i] = list(comp)

            cluster_assignments = {
                record_id : cluster_id
                for cluster_id, record_ids in clusters.items()
                for record_id in record_ids
            }

            original_cluster_ids = None

        # add in the singletons
        if len(clusters) == 0:
            i = -1

        cluster_id = i + 1
        logging.debug(f"new cluster_id for singletons starts from: {cluster_id}")

        an_df = an_df.reset_index()
        an_df = an_df[an_df.record_id.isin(cluster_assignments.keys()) == False].reset_index(drop=True)
        an_df['cluster_id'] = an_df.index + cluster_id
        # NOTE: next 4 lines are for a 1000x speedup over pandas groupby
        keys, values = an_df[['cluster_id', 'record_id']].values.T
        ukeys, index = np.unique(keys, True)
        arrays = np.split(values, index[1:])
        singleton_clusters = dict(zip(ukeys, [list(a) for a in arrays]))
        clusters.update(singleton_clusters)
        cluster_assignments.update(an_df.set_index('record_id').to_dict()['cluster_id'])
        self.stats_dict['n_initial_clusters'] = len(clusters)
        del must_links_df
        gc.collect()
        return clusters, cluster_assignments, original_cluster_ids

    # @log_runtime_and_memory
    def save_df_to_disk(self, df):
        table = pa.Table.from_pandas(df)
        pq.write_table(table, self.edges)

    # @log_runtime_and_memory
    @profile
    def get_potential_edges(self, potential_edges_files, flipped0_edges_file, gt_1s_df, cluster_logic, cluster_info, uid_cols, eid_col, **kw):
        '''
        Use all predictions file to make a list of edges that the constrained
        clustering algorithm should try to add.

        Args:
            potential_edges_files (list): paths to the potential links files
            flipped0_edges_file (str): path to the flipped0-links file
            gt_1s_df (pd.DataFrame): for incremental runs, subset of the must-link df that are "new" 1s
            cluster_logic (module): user-defined constraint functions
            cluster_info (pd.DataFrame): all-names file, with only the columns relevant for clustering
            uid_cols (list): all-name columns with compare_type UniqueID
            eid_col (str): all-name column with compare_type ExistingID

        '''
        total_auto_invalid_edges = 0
        total_invalid_edges = 0
        for i, potential_edges_file in enumerate(potential_edges_files):

            potential_edges_df = load_parquet(potential_edges_file, conditions_dict={'potential_edge': 1})
            potential_edges_df['gt'] = 0

            if i == 0: # for first iteration only
                if (gt_1s_df is not None) and (os.path.exists(flipped0_edges_file)):
                    flipped0_edges_file = pd.read_csv(flipped0_edges_file, low_memory=False)
                    flipped0_edges_file['gt'] = 0
                    potential_edges_df = pd.concat([potential_edges_df, flipped0_edges_file], join='inner')
                    potential_edges_df = potential_edges_df.reset_index(drop=True)

            # get clustering_phat
            potential_edges_df['phat'] = -1
            for model_name in potential_edges_df.model_to_use.unique():
                potential_edges_df.loc[potential_edges_df.model_to_use == model_name, 'phat'] = \
                        potential_edges_df['%s_match_phat' % model_name]

            cols = ['record_id_1', 'record_id_2', 'gt', 'phat']
            potential_edges_df = potential_edges_df[cols]

            if i == 0: # for first iteration only
                if gt_1s_df is not None:
                    gt_1s_df['gt'] = 1
                    gt_1s_df['phat'] = 1
                    potential_edges_df = pd.concat([potential_edges_df, gt_1s_df[cols]])

            potential_edges_df = potential_edges_df.reset_index(drop=True)

            logger.trace('Dropping invalid edges')
            potential_edges_df = pd.merge(potential_edges_df, cluster_info, left_on='record_id_1', right_index=True)
            potential_edges_df = pd.merge(potential_edges_df, cluster_info, left_on='record_id_2', right_index=True, suffixes=['_1', '_2'])
            potential_edges_df = potential_edges_df.reset_index(drop=True)

            starting_cols = potential_edges_df.columns.tolist()
            n_before_drop_auto_invalid_edges = len(potential_edges_df)
            potential_edges_df = self.auto_is_valid_edge(
                    potential_edges_df, uid_cols,
                    self.params.allow_clusters_w_multiple_unique_ids, self.params.leven_thresh, eid_col)
            potential_edges_df = potential_edges_df[starting_cols]
            n_before_drop_invalid_edges = len(potential_edges_df)
            potential_edges_df['valid'] = cluster_logic.is_valid_link(potential_edges_df)
            potential_edges_df = potential_edges_df[
                (potential_edges_df.gt == 1) | (potential_edges_df.valid == 1)]
            potential_edges_df = potential_edges_df[starting_cols]
            auto_invalid_edges = n_before_drop_auto_invalid_edges - n_before_drop_invalid_edges
            invalid_edges = n_before_drop_invalid_edges - len(potential_edges_df)

            total_auto_invalid_edges += auto_invalid_edges
            total_invalid_edges += invalid_edges

            if i == 0:
                valid_potential_edges_df = potential_edges_df.copy()
            else:
                valid_potential_edges_df = pd.concat([valid_potential_edges_df, potential_edges_df])

            del potential_edges_df
            gc.collect()
        logger.info(f"Auto-invalid links: {total_auto_invalid_edges}")
        logger.info(f"Invalid links: {total_invalid_edges}")
        self.stats_dict['n_invalid_links'] = total_invalid_edges

        valid_potential_edges_df = valid_potential_edges_df.reset_index(drop=True)
        valid_potential_edges_df = valid_potential_edges_df.sort_values(
                ['gt', 'phat'], ascending=[False, False])

        logger.trace('Applying link priority.')
        valid_potential_edges_df['original_order'] = valid_potential_edges_df.index
        valid_potential_edges_df = cluster_logic.apply_link_priority(valid_potential_edges_df)

        self.stats_dict['n_potential_edges'] = len(valid_potential_edges_df)
        # NOTE: takes a little while, but worth it later to avoid iterrows

        logger.trace('Saving valid_links_df to disk as parquet file')

        self.save_df_to_disk(valid_potential_edges_df)
        del valid_potential_edges_df
        gc.collect()

    # @log_runtime_and_memory
    @profile
    def load_cluster_info(self, all_names_file, uid_cols, eid_col, cluster_logic, **kw):
        '''Read in the all_names information needed for cluster constraint checking. Columns
        defined in the config as compare type UniqueID or ExistingID will automatically be loaded
        (as strings, with missing values represented as NA). Other columns you wish to be loaded
        should be defined in the user-defined `get_columns_used()` function.

        Args:
            all_names_file (str): path to the all-names file
            uid_cols (list): all-name columns with compare_type UniqueID
            eid_col (str): all-name column with compare_type ExistingID
            cluster_logic (module): user-defined constraint functions

        Returns:
            pd.DataFrame: all-names file, with only the columns relevant for clustering

            ===========================  =======================================================
            record_id                    unique record identifier
            <uid column(s)>              columns with compare_type UniqueID
            <eid column(s)>              columns with compare_type ExistingID
            <user-constraint column(s)>  (optional) columns mentioned in `get_columns_used()`
            ===========================  =======================================================
        '''

        # add things that are missing from get_columns_used

        cols_needed = cluster_logic.get_columns_used()
        if cols_needed == 'all':
            column_dtypes = {field:'object' for field in self.schema.variables.get_an_column_names()}
        else: 
            column_dtypes = cols_needed.copy()
            if ("record_id" not in column_dtypes):
                column_dtypes["record_id"] = 'object'
            if ("dataset" not in column_dtypes):
                column_dtypes["dataset"] = 'object'

        id_cols_to_load_as_obj = uid_cols
        if eid_col is not None:
            id_cols_to_load_as_obj.append(eid_col)

        for col in id_cols_to_load_as_obj:
            if col in column_dtypes:
                if column_dtypes[col] != 'object':
                    logger.warning(f"Changing dtype for {col} field to object.")
                    column_dtypes[col] = 'object'
                logger.warning(f"Note, missing values in {col} field will not be "
                               f"filled with '', since this is an ID field.")
            else:
                column_dtypes[col] = 'object'

        table = pq.read_table(all_names_file)
        cluster_info =  table.to_pandas()
        cluster_info = cluster_info[list(column_dtypes.keys())]
        del column_dtypes['record_id']
        cluster_info = cluster_info.set_index('record_id')
        # change the dtypes in all_names based on user-defined dtypes in constraints
        for col, col_dtype in column_dtypes.items():
            if col_dtype in ['int', 'float']:
                cluster_info[col] = pd.to_numeric(cluster_info[col])
            elif col_dtype in ['date']:
                cluster_info[col] = pd.to_datetime(cluster_info[col])
            elif col_dtype in ['str', 'string', 'object']:
                cluster_info[col] = cluster_info[col].fillna("")
                cluster_info[col] = cluster_info[col].astype(str)

        # replace empty string with NaN for uids and eid
        for col in id_cols_to_load_as_obj:
            cluster_info[col] = cluster_info[col].replace('', np.nan)
        return cluster_info

    # @log_runtime_and_memory
    def get_ci_ix_map(self, cluster_info):
        ci_ix_map = dict(zip(cluster_info.index.tolist(), range(len(cluster_info))))
        return ci_ix_map

    # @log_runtime_and_memory
    @profile
    def cluster_potential_edges(self, clusters, cluster_assignments, original_cluster_ids,
                cluster_info, cluster_logic, uid_cols, eid_col, **kw):
        '''For clusters by add potential edges to the cluster graph in order of importance, skipping those
        that cause violations.

        Args:
            clusters (dict): maps a cluster id to a list of record ids -- post initialization
            cluster_assignments (dict): maps a record_id to a cluster_id -- post initialization
            original_cluster_ids (set): set: cluster ids that are already in use (only for incremental)
            cluster_info (pd.DataFrame): all-names file, with only the columns relevant for clustering

                ===========================  =======================================================
                record_id                    unique record identifier
                <uid column(s)>              columns with compare_type UniqueID
                <eid column(s)>              columns with compare_type ExistingID
                <user-constraint column(s)>  (optional) columns mentioned in `get_columns_used()`
                ===========================  =======================================================

            potential_edges (deque): each element is a dict version of a potential edge's record
            cluster_logic (module): user-defined constraint functions
            uid_cols (list): all-name columns with compare_type UniqueID
            eid_col (str): all-name column with compare_type ExistingID

        Returns:
            dict: maps record_id to cluster_id
        '''
        # track things
        invalid_edges = 0
        invalid_clusters = 0
        merges = 0
        pf = pq.ParquetFile(self.edges)
        nrows = pf.metadata.num_rows
        logger.info(f"total number of edges: {nrows}")

        # create ix map so can use faster iloc as opposed to loc
        ci_ix_map = self.get_ci_ix_map(cluster_info)
        i = 0
        logger.debug(f"batch size: {self.params.cluster_batch_size}")
        for edge_pf in pf.iter_batches(batch_size=self.params.cluster_batch_size, use_threads=True):
            for edge in edge_pf.to_pylist():
                if (self.params.verbose is not None) and (i % self.params.verbose == 0):
                    logger.info(f"  Checked {i} of {nrows} edges: {invalid_clusters} invalid clusters, {merges} merges.")

                record_id_1 = edge["record_id_1"]
                record_id_2 = edge["record_id_2"]

                edge_is_gt = (edge["gt"] == 1)

                cluster_id_1 = cluster_assignments[record_id_1]
                cluster_id_2 = cluster_assignments[record_id_2]


                if (cluster_id_1 != cluster_id_2):
                    cluster_1 = clusters[cluster_id_1]
                    cluster_2 = clusters[cluster_id_2]
                    new_cluster = cluster_1 + cluster_2

                    if len(new_cluster) > 2:
                        new_cluster_info = cluster_info.iloc[[ci_ix_map[ncr] for ncr in new_cluster]].copy()

                        new_cluster_info['cluster'] = \
                                new_cluster_info.index.isin(cluster_1).astype(int) # differentiate

                        new_cluster_info['new_edge'] = \
                                new_cluster_info.index.isin([record_id_1, record_id_2])

                        # NOTE: new_edge is a way of checking an edge constraint within
                        #       is_valid_cluster (in case you want to enforce an edge
                        #       constraint, but only for clusters that meet some criteria)

                    # if len of new_cluster is 2 then in theory we already vetted by checking the edge
                    # (assuming that all cluster constraints were put into the edge constraint list if possible)

                    if len(new_cluster) > 2:
                        cluster_auto_valid = self.auto_is_valid_cluster(
                            new_cluster_info, uid_cols,
                            self.params.allow_clusters_w_multiple_unique_ids, self.params.leven_thresh, eid_col)

                    if  (len(new_cluster) == 2) or \
                        (cluster_auto_valid and (edge_is_gt or cluster_logic.is_valid_cluster(new_cluster_info, edge['phat']))):

                        new_cluster_id = min(cluster_id_1, cluster_id_2)

                        if original_cluster_ids is not None:
                            if cluster_id_1 in original_cluster_ids:
                                new_cluster_id = cluster_id_1
                            elif cluster_id_2 in original_cluster_ids:
                                new_cluster_id = cluster_id_2

                        clusters.pop(cluster_id_1)
                        clusters.pop(cluster_id_2)
                        clusters[new_cluster_id] = new_cluster

                        for record_id in new_cluster:
                            cluster_assignments[record_id] = new_cluster_id

                        merges += 1

                    else:
                        invalid_clusters += 1

                i += 1

        logger.info(f"Invalid clusters: {invalid_clusters}")
        self.stats_dict['n_invalid_clusters'] = invalid_clusters
        logger.info(f"n_merges: {merges}")
        n_clusters = len(clusters)
        logger.info(f"Number of clusters total: {n_clusters}")
        self.stats_dict['n_clusters'] = n_clusters
        n_singleton_clusters = len([recs for c_id, recs in clusters.items() if (len(recs) == 1)])
        logger.info(f"Number of singleton clusters: {n_singleton_clusters}")
        self.stats_dict['n_singleton_clusters'] = n_singleton_clusters

        cluster_assignments = {k: str(v) for k, v in cluster_assignments.items()}
        return cluster_assignments




import numpy as np
import pandas as pd
from sklearn import metrics

import logging
logger = logging.getLogger()


def get_precision(eval_df, thresh):
    '''Compute precision.
 
    Args:
        eval_df (pd.DataFrame): contains three columns needed for performance metrics: phat, outcome, weight
        thresh (float): probability threshold that optimizes fscore

    Returns:
        float: precision
    '''

    predicted_positives = eval_df[eval_df.phat >= thresh].copy()
    if len(predicted_positives) == 0:
        return None
    precision = np.average(predicted_positives.outcome, weights=predicted_positives.weight)
    return precision


def get_recall(eval_df, thresh):
    '''Compute recall.

    Args:
        eval_df (pd.DataFrame): contains three columns needed for performance metrics: phat, outcome, weight
        thresh (float): probability threshold that optimizes fscore

    Returns:
        float: recall
    '''

    actual_positives = eval_df[eval_df.outcome == 1].copy()
    actual_positives['yhat'] = actual_positives.phat >= thresh
    if len(actual_positives) == 0:
        return None
    recall = np.average(actual_positives.yhat, weights=actual_positives.weight)

    return recall


def get_fscore(eval_df, thresh, beta):
    '''Compute fscore.

    Args:
        eval_df (pd.DataFrame): contains three columns needed for performance metrics: phat, outcome, weight
        thresh (float): probability threshold that optimizes fscore
        beta (float): ratio of recall weighting to precision weighting (e.g. 0.5 weights precision double)

    Returns:
        float: fscore
    '''

    precision = get_precision(eval_df, thresh)
    recall = get_recall(eval_df, thresh)
    fscore = None
    if precision and recall:
        fscore = (1 + (beta**2)) * ((precision * recall) / (((beta**2) * precision) + recall))

    return fscore


def get_accuracy(eval_df, thresh):
    '''Compute accuracy.

    Args:
        eval_df (pd.DataFrame): contains three columns needed for performance metrics: phat, outcome, weight
        thresh (float): probability threshold that optimizes fscore

    Returns:
        float: accuracy
    '''

    eval_df['correct'] = eval_df.outcome == (eval_df.phat >= thresh).astype(int)
    if len(eval_df) == 0:
        return None
    accuracy = np.average(eval_df.correct, weights=eval_df.weight)

    return accuracy


def get_fpr(eval_df, thresh):
    '''Compute false postive rate.

    Args:
        eval_df (pd.DataFrame): contains three columns needed for performance metrics: phat, outcome, weight
        thresh (float): probability threshold that optimizes fscore

    Returns:
        float: false positive rate
    '''

    actual_negatives = eval_df[eval_df.outcome == 0].copy()
    actual_negatives['yhat'] = actual_negatives.phat >= thresh
    if len(actual_negatives) == 0:
        return None
    fpr = np.average(actual_negatives.yhat, weights=actual_negatives.weight)

    return fpr


def pairwise_metrics(labeled_preds, threshold, phat_col, outcome, fscore_beta, weight):
    '''Generate evaluation metrics that are standard for classification problems (e.g. precision, auc).

    Args:
        labeled_preds (pd.DataFrame): df with labeled test phats and columns needed for evaluation (limited to a universe)
        threshold (float): probability threshold that optimizes fscore
        phat_col (str): phat column for evaluation
        outcome (str): outcome to evaluate
        fscore_beta (float): ratio of recall weighting to precision weighting (e.g. 0.5 weights precision double)
        weight (bool): should the evaluation metrics utilize the selection model weights

    Return:
        list of floats: base_rate, precision, recall, fpr, fnr, auc, accuracy, fscore
    '''
    labeled_preds = labeled_preds.copy()

    preds = labeled_preds[phat_col]
    labels = labeled_preds[outcome]

    if weight:
        weights = labeled_preds[phat_col.replace('_match_phat', '_weight')]
    else: 
        weights = [1 for i in np.arange(len(labeled_preds))]

    base_rate = np.average(labels, weights=weights)
    auc = metrics.roc_auc_score(labels, preds, sample_weight=weights)

    eval_df = pd.DataFrame(data={
        'phat':preds, 
        'outcome':labels,
        'weight':weights
    })

    precision = get_precision(eval_df, threshold)
    recall = get_recall(eval_df, threshold)
    fpr = get_fpr(eval_df, threshold)
    fnr = 1 - recall
    fscore = get_fscore(eval_df, threshold, fscore_beta)
    accuracy = get_accuracy(eval_df, threshold)

    return base_rate, precision, recall, fpr, fnr, auc, accuracy, fscore


def find_best_threshold(df, beta, weight):
    '''Find the threshold that optimizes fscore.

    Args:
        df (pd.DataFrame): predicted and actual values
            =====================   =======================================================
            outcome                 whether the pair is a match or not
            phat                    predicted probability of match
            weight                  weight to use for evaluation 
            =====================   =======================================================
        beta (float): ratio of recall weighting to precision weighting (e.g. 0.5 weights precision double)
        weight (bool): should we use the selection model weights when picking a threshold?

    Returns:
        float: probability threshold that optimizes fscore
    '''

    df = df.copy()

    if not weight:
        df['weight'] = 1

    logger.info('Finding optimal threshold.')

    best_t = -1
    best_fscore = -1
    for t in range(99, 1, -1):
        try:
            fscore = get_fscore(df[['phat', 'outcome', 'weight']], t/100.0, beta)
        except:
            continue
        if (fscore is not None) and (fscore > best_fscore):
            best_fscore = fscore
            best_t = t / 100.0

    if best_t == -1:
        logger.error('Problem calculating optimal threshold.')
        raise ValueError

    return best_t

def get_cv_metrics(mod): 
    '''Create table of cv scoring metrics between train and test.

    Args: 
        mod (sklearn model): Fit gridsearch object

    Returns: 
        pd.DataFrame: train/test performance for each param set in grid

    '''

    cv_results = pd.DataFrame.from_records(mod.cv_results_['params'])

    cv_results['train_score'] = mod.cv_results_['mean_train_score']
    cv_results['test_score'] = mod.cv_results_['mean_test_score']
    cv_results['train_test_difference'] = \
            cv_results['train_score'] - cv_results['test_score']

    return cv_results


def evaluate_predictions(phat_df, model_type, phat_col, outcome, weight=False,
        default_threshold=0.5, missingness_model_threshold_boost=0.2, 
        optimize_threshold=False, fscore_beta=1.0):
    '''Calculates metrics such as precision, recall, fscore, etc. for the pairwise record
    match predictions. Also, get the threshold that maximizes f score.

    Args:
        phat_df (pd.DataFrame): df with labeled test phats and columns needed for evaluation
        model_type (str): match or selection
        phat_col (str): phat column for evaluation
        outcome (str): outcome to evaluate
        weight (bool): should the evaluation metrics utilize the selection model weights
        default_threshold (float): threshold for match/non-match (use if don't find optimal)
        missingness_model_threshold_boost (float): value to add to default threshold if missingess model
        optimize_threshold (bool): should we find the threshold that optimizes f1
        fscore_beta (float): ratio of recall weighting to precision weighting (e.g. 0.5 weights precision double)

    Returns:
        float: probability threshold that optimizes fscore
        dict: keys are universe str, values are dicts with perf metrics
    '''
    phat_df = phat_df.copy()

    all_model_stats = {}

    # hacky for now
    if 'no_' in phat_col and 'basic' not in phat_col:
        default_threshold = default_threshold + missingness_model_threshold_boost

    weight_col = phat_col.replace(f'{model_type}_phat', 'weight')
    if not weight:
        phat_df[weight_col] = 1
    
    if outcome != 'match_train_eligible':
        phat_df = phat_df[phat_df.match_train_eligible == 1]

    if len(phat_df) == 0:
        logger.info(f"Unable to measure performance (no labeled testing data available).")
        return default_threshold, all_model_stats

    logger.info(f'Number of labeled test rows evaluated: {len(phat_df)}')

    if optimize_threshold:
        try:
            df = phat_df[[phat_col, outcome, weight_col]].copy()
            df.columns = ['phat', 'outcome', 'weight']
            threshold = find_best_threshold(df, fscore_beta, weight)
        except:
            threshold = default_threshold
            logger.warning(f'Could not find optimal threshold, using {threshold}.')
    else:
        threshold = default_threshold
    logger.info(f"Threshold: {threshold}")

    for universe in ['all pairs', 'non exactmatch pairs', 'exactmatch pairs']:

        model_stats = {}

        phat_univ_df = phat_df.copy()
        if universe == 'non exactmatch pairs':
            phat_univ_df = phat_df[phat_df.exactmatch == 0].copy()
        elif universe == 'exactmatch pairs':
            phat_univ_df = phat_df[phat_df.exactmatch == 1].copy()

        # log phat distributions
        try:

            one_phats = phat_univ_df[phat_univ_df[outcome] == 1][phat_col]
            one_phat_dist = pd.value_counts(pd.cut(one_phats, np.arange(0, 1.1, .1)), normalize=True, sort=False)
            logger.trace(f'Phat distribution of actual 1s ({universe}): \n{one_phat_dist.to_string()}')
            model_stats['phat_distribution_1s'] = list(one_phat_dist)

            zero_phats = phat_univ_df[phat_univ_df[outcome] == 0][phat_col]
            zero_phat_dist = pd.value_counts(pd.cut(zero_phats, np.arange(0, 1.1, .1)), normalize=True, sort=False)
            logger.trace(f'Phat distribution of actual 0s ({universe}): \n{zero_phat_dist.to_string()}')
            model_stats['phat_distribution_0s'] = list(zero_phat_dist)

        except:
            logger.info(f"Issue calculating phat distributions ({universe}).")

        try:
            if len(phat_univ_df) == 0:
                baserate = precision = recall = fpr = fnr = auc = accuracy = fscore = None
            else:
                baserate, precision, recall, fpr, fnr, auc, accuracy, fscore = pairwise_metrics(
                    phat_univ_df, threshold, phat_col, outcome, fscore_beta, weight)

            logger.info(f"Base rate ({universe}): {baserate}")
            model_stats['baserate'] = float(baserate) if baserate else None

            logger.info(f"Precision ({universe}): {precision}")
            model_stats['precision'] = float(precision) if precision else None

            logger.info(f"Recall ({universe}): {recall}")
            model_stats['recall'] = float(recall) if recall else None

            logger.info(f"False positive rate ({universe}): {fpr}")
            model_stats['fp_rate'] = float(fpr) if fpr else None

            logger.info(f"False negative rate ({universe}): {fnr}")
            model_stats['fn_rate'] = float(fnr) if fnr else None

            logger.info(f"AUC ({universe}): {auc}")
            model_stats['auc'] = float(auc) if auc else None

            logger.info(f"F-score ({universe}): {fscore}")
            model_stats['fscore'] = float(fscore) if fscore else None

        except:
            message = f"Issue with given threshold -- not all areas of confusion matrix present ({universe})."
            if universe != 'exactmatch pairs':
                logger.warning(message)
            else: 
                logger.debug(message)

        all_model_stats[universe] = model_stats

    return threshold, all_model_stats


def evaluate_models(phats_df, outcome, model_type, weight=False, default_threshold=0.5, 
            missingness_model_threshold_boost=0.2, optimize_threshold=False, fscore_beta=1.0, 
            stats_dict=None):
    '''Wrapper for evaluating different models (e.g. basic and no-dob) on different universes.

    Args:
        phat_df (pd.DataFrame): df with labeled test phats and columns needed for evaluation
        outcome (str): outcome to evaluate
        model_type (str): either "selection" or "match"
        weight (bool): should the evaluation metrics utilize the selection model weights
        default_threshold (float): threshold for match/non-match (use if don't find optimal)
        missingness_model_threshold_boost (float): value to add to default threshold if missingess model (use if don't find optimal)
        optimize_threshold (bool): should we find the threshold that optimizes fscore
        fscore_beta (float): ratio of recall weighting to precision weighting (e.g. 0.5 weights precision double)

    Return:
        dict: maps model name (e.g. basic, no-dob) to thresholds (no return type)
    '''
    model_names = [col.replace(f'_{model_type}_phat', '')
                   for col in phats_df.columns.tolist()
                   if f'_{model_type}_phat' in col]

    if model_type == 'selection':
        optimize_threshold = False

    thresholds = {}
    model_type_model_stats = {}
    for model_name in model_names:

        phats_to_eval_df = phats_df[phats_df.model_to_use == model_name]
        # NOTE: will cause problems if this universe isn't represented in the labeled data;
        #       use basic poplation as a backup
        if len(phats_to_eval_df) == 0:
            phats_to_eval_df = phats_df[phats_df.model_to_use == 'basic']

        phat_col = f'{model_name}_{model_type}_phat'

        logger.info(f'----- EVALUATING {model_name.upper()} {model_type.upper()} MODEL -----')
        thresholds[model_name], \
        model_type_model_stats[model_name] = evaluate_predictions(
                phats_to_eval_df, model_type, phat_col, outcome, weight,
                default_threshold, missingness_model_threshold_boost, optimize_threshold, fscore_beta)
    
    stats_dict[f"model_stats__{model_type}"] = model_type_model_stats
    if model_type == 'match':
        stats_dict[f"model_thresholds__{model_type}"] = thresholds

    if model_type == 'match':
        return thresholds



from unittest.mock import patch
import warnings

import numpy as np

from namematch.generate_data_rows import GenerateDataRows
from namematch.comparison_functions import (
    get_name_probabilities,
    try_switch_first_last_name,
    compare_strings,
    compare_numbers,
    compare_categories,
    compare_dates,
    compare_geographies,
    generate_label,
)


def test_get_name_probabilities(an_df, side_by_side_df, params_and_schema, logger_for_testing):
    params, schema = params_and_schema

    # too many warnings
    warnings.filterwarnings("ignore")

    first_name_col = "first_name"
    last_name_col = "last_name"

    with patch('namematch.generate_data_rows.logger', logger_for_testing) as mock_debug:
        generate_data_rows = GenerateDataRows(
            params,
            schema,
            all_names_file=an_df,
            candidate_pairs_file=None,
            output_dir=None,
        )

        np_object = generate_data_rows.generate_name_probabilities_object(an_df, first_name_col, last_name_col)

    # get probabilities
    probs = get_name_probabilities(side_by_side_df, np_object, first_name_col, last_name_col)

    # test
    expected = [
        "prob_name_1",
        "prob_name_2",
        "prob_rev_name_1",
        "prob_rev_name_2",
        "count_pctl_name_1",
        "count_pctl_name_2",
        "count_pctl_fn_1",
        "count_pctl_fn_2",
        "count_pctl_ln_1",
        "count_pctl_ln_2",
        "prob_same_name",
        "prob_same_name_rev_1",
        "prob_same_name_rev_2"
        ]
    assert list(side_by_side_df) + expected == list(probs)

    for i, row in probs.iterrows():

        # this is true of the artificial data that is being used in this test
        # obviously it won't always be true

        if (row.first_name_1.startswith("J")):
            assert row.prob_rev_name_1 < row.prob_name_1
        else:
            assert row.prob_rev_name_1 > row.prob_name_1

        if (row.first_name_2.startswith("J")):
            assert row.prob_rev_name_2 < row.prob_name_2
        else:
            assert row.prob_rev_name_2 > row.prob_name_2


def test_try_switch_first_last_name(side_by_side_df_with_probs):
    # switch
    switched = try_switch_first_last_name(side_by_side_df_with_probs, "first_name", "last_name")

    # test
    for i, row in switched.iterrows():
        assert row.first_name_1.startswith("J")
        assert row.first_name_2.startswith("J")
        assert row.last_name_1.startswith("D")
        assert row.last_name_2.startswith("D")


def test_compare_strings(side_by_side_df):
    varname = "string"

    # compare
    compared = compare_strings(side_by_side_df, varname)

    # test
    assert 0 == compared.string_missing.sum()
    assert side_by_side_df.shape[0], compared.shape[0]
    assert (side_by_side_df.string_1 == side_by_side_df.string_2).sum() == compared.string_exact_match.sum()
    assert compared.string_exact_match.sum() < compared.string_soundex.sum()


def test_compare_numbers(side_by_side_df):
    df = side_by_side_df.astype({"number_1": str,"number_2": str})
    varname = "number"

    # compare
    compared = compare_numbers(df, varname)

    # test
    assert 0 == compared.number_missing.sum()
    assert df.shape[0] == compared.shape[0]


def test_compare_categories(side_by_side_df):
    side_by_side_df = side_by_side_df.fillna('')
    varname = "cat"

    # compare
    compared = compare_categories(side_by_side_df, varname)

    # test
    assert side_by_side_df.shape[0] == compared.shape[0]
    assert 0 == compared.cat_missing.isna().sum()
    assert 0 != compared.cat_exact_match.isna().sum()


def test_compare_dates(side_by_side_df):
    varname = "date"

    # compare
    compared = compare_dates(side_by_side_df, varname)

    # test
    assert side_by_side_df.shape[0] == compared.shape[0]
    assert 0 == compared.date_missing.isna().sum()
    assert 0 == compared.date_missing.sum()
    assert 0 == (compared.date_day_diff > 365).sum()


def test_compare_geographies(side_by_side_df):
    varname = "geo"

    # compare
    compared = compare_geographies(side_by_side_df, varname)

    # test
    assert side_by_side_df.shape[0] == compared.shape[0]
    assert 0 == compared.geo_missing.isna().sum()
    assert 0 == compared.geo_missing.sum()


def test_generate_label(ids_df):
    # the data is laid out in the following way:
    # uniqueid         ID        uid
    #    match      match      match
    #    match      match    missing
    #    match      match   mismatch
    #    match      match  nearmatch
    #    match  nearmatch   mismatch
    #  missing    missing  nearmatch

    uid_vars = ['uid']

    # test with no leven_thresh
    labels = generate_label(ids_df, uid_vars, leven_thresh=None)

    np.testing.assert_array_equal(['1', '', '0', '0', '0', '0'], labels.tolist())

    # test with leven_thresh
    labels = generate_label(ids_df, uid_vars, leven_thresh=1)
    np.testing.assert_array_equal(['1', '', '0', '', '0', ''], labels.tolist())




import csv
import logging
import numpy as np
import os
import pandas as pd
import pickle

import pyarrow as pa
import pyarrow.parquet as pq

from datetime import datetime

from namematch.base import NamematchBase
from namematch.data_structures.parameters import Parameters
from namematch.data_structures.schema import Schema
from namematch.utils.utils import *

logger = logging.getLogger(__name__)

class GenerateOutput(NamematchBase):
    '''
    Args:
        params (Parameters object): contains parameter values
        schema (Schema object): contains match schema info (files to match, variables to use, etc.)
        all_names_file (str): path to the all-names file
        cluster_assignments_file (str): path to the cluster-assignments file
        an_output_file (str): path to the all-names-with-clusterid file
        output_dir (str): path to final output directory
    '''
    def __init__(
        self,
        params,
        schema,
        all_names_file,
        cluster_assignments_file,
        an_output_file,
        output_dir,
        output_file_uuid=None,
        *args,
        **kwargs
    ):
        super(GenerateOutput, self).__init__(params, schema, *args, **kwargs)

        self.all_names_file = all_names_file
        self.cluster_assignments_file = cluster_assignments_file
        self.an_output_file = an_output_file
        self.output_dir = output_dir
        self.output_file_uuid = output_file_uuid

    @property
    def output_files(self):
        output_files = [self.an_output_file]
        for data_file in self.schema.data_files.data_files:
            if self.output_file_uuid:
                self.output_file_name = data_file.output_file_stem + f"_with_clusterid_{self.output_file_uuid}.csv"
            else:
                output_file_name = data_file.output_file_stem + "_with_clusterid.csv"

            output_files.append(os.path.join(self.output_dir, output_file_name))

        return output_files

    def main(self, **kw):
        '''Read in the cluster assignments dictionary and use it to create all-names-with-cluster-id
        and the "with-cluster-id" versions of input dataset.
        '''
        with open(self.cluster_assignments_file, "rb") as f:
            cluster_assignments = pickle.load(f)

        an_df = self.create_allnames_clusterid_file(
                self.all_names_file, cluster_assignments,
                self.params.blocking_scheme['cosine_distance']['variables'])
        an_df.to_csv(self.an_output_file, index=False)

        self.output_clusterid_files(
            self.schema.data_files.data_files,
            cluster_assignments,
            self.output_dir,
            self.output_file_uuid
        )

        self.stats_dict['end'] = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")

    # @log_runtime_and_memory
    def create_allnames_clusterid_file(self, all_names_file, cluster_assignments, cleaned_col_names, **kw):
        '''Create all-names-with-clusterid dataframe.

        Args:
            all_names_file (str): path to the all-names file
            cluster_assignments (dict): maps record_id to cluster_id
            cleaned_col_names (list): all-name columns used in cosine blocking

        Returns:
            pd.DataFrame: all-names-with-cluster-id

            =====================   =======================================================
            record_id               unique record identifier
            file_type               either "new" or "existing"
            <fields for matching>   both for the matching model and for constraint checking
            blockstring             concatenated version of blocking columns (sep by ::)
            drop_from_nm            flag, 1 if met any "to drop" criteria 0 otherwise
            cluster_id              unique person identifier, no missing values
            =====================   =======================================================
        '''

        logger.info('Generating all_names file with cluster_id.')

        table = pq.read_table(all_names_file)
        all_names = table.to_pandas()
        all_names["cluster_id"] = all_names["record_id"].map(cluster_assignments)
        for cleaned_col in cleaned_col_names:
            all_names[cleaned_col] = all_names[f'tmp_raw__{cleaned_col}']
        all_names = all_names.drop(columns=[f'tmp_raw__{col}' for col in cleaned_col_names])

        return all_names

    # @log_runtime_and_memory
    def output_clusterid_files(self, data_files, cluster_assignments, output_dir, output_file_uuid=None, **kw):
        '''For each input file, construct a matching output file that has the
        cluster_id column, and write it.

        Args:
            data_files (list of DataFile objects): contains info about each input file
            cluster_assignments (dict): maps record_id to cluster_id
            output_dir (str): the path that was supplied when the name match object was created
        '''
        if output_file_uuid is None:
            output_file_uuid = self.output_file_uuid

        # remove output files if they exist (necessary for
        # input_files that have the same output_file_stem)
        for data_file in data_files:
            if output_file_uuid:
                output_file_name = data_file.output_file_stem + f"_with_clusterid_{output_file_uuid}.csv"
            else:
                output_file_name = data_file.output_file_stem + "_with_clusterid.csv"
            output_file_name = os.path.join(output_dir, output_file_name)
            if os.path.exists(output_file_name):
                os.remove(output_file_name)

        # update each of the input files
        for data_file in data_files:

            if output_file_uuid:
                output_file_name = data_file.output_file_stem + f"_with_clusterid_{output_file_uuid}.csv"
            else:
                output_file_name = data_file.output_file_stem + "_with_clusterid.csv"

            output_file_path = os.path.join(output_dir, output_file_name)
            if not os.path.exists(output_file_path):
                # read input file
                df = pd.read_csv(data_file.filepath, dtype=object, encoding="ISO-8859-1")
            else:
                # read existing output file
                df = pd.read_csv(output_file_path, dtype=object, encoding="ISO-8859-1")

            logger.info(f"Writing {output_file_name} file.")

            df['temp_nm_rec_id'] = create_nm_record_id(
                    data_file.nickname, df[data_file.record_id_col])

            cluster_id_col = f"{data_file.cluster_type}_id"
            df[cluster_id_col] = df['temp_nm_rec_id'].map(cluster_assignments)

            n_missing = df[cluster_id_col].isnull().sum()
            if  n_missing > 0:
                logger.warning(f"{n_missing} rows without cluster_ids (this shouldn't happen)")
                df = df[df[cluster_id_col].isnull() == False]

            df.drop(columns=['temp_nm_rec_id']).to_csv(
                    output_file_path, index=False, quoting=csv.QUOTE_NONNUMERIC)




import Levenshtein as edist
import numpy as np
from collections import defaultdict

def _editCounts(name_samp):
    # to compute probability of edit operations use a subsample of names
    edit_count = defaultdict(int)
    p = len(name_samp)
    total_edits = 0
    for i in range(p):
        for j in range(i + 1, p):
            if i < j:
                edits = edist.editops(name_samp[i], name_samp[j])
                p = len(edits)
                lene = p
                total_edits += len(edits)
                for k in range(lene):
                    edit_count[edits[k]] += 1
    return edit_count, total_edits


def _ngramCount(name_list, ngram_len):
    ngram_count = defaultdict(int)
    for i in range(len(name_list)):
        current_name = name_list[i]
        if len(current_name) > ngram_len - 1:
            for start in range(len(current_name) - (ngram_len - 1)):
                ngram_count[current_name[start:(start + ngram_len)]] += 1
                ngram_count[current_name[start:((start + ngram_len)-1)]] += 1
            ngram_count[current_name[-(ngram_len - 1):]] += 1
    return ngram_count


def _probName(name, ngram_len, ngram_count, smoothing, memoize):
    log_prob = 0.0
    for start in range(len(name) - (ngram_len - 1)):
        numer = ngram_count[name[start:(start + ngram_len)]] + smoothing
        denom = ngram_count[name[start:(start + ngram_len)-1]] + smoothing
        if not denom:
            denom += .000001 # avoid div by zero error
        log_prob += np.log(numer / denom)
    memoize[name] = np.exp(log_prob)
    return memoize


def _condProbName(name1, name2, edit_count, total_edits, smoothing, cp_memoize):
    # computes the conditional probability of arriving at name1
    # by performing a series of operation on name2.
    temp_count = defaultdict(float)
    holder = 0.0
    for k, v in edit_count.items():
        temp_count[k] = v / total_edits
    edits = edist.editops(name1, name2)
    for e in edits:
        holder += np.log(temp_count[e] + smoothing)
    log_cnd_prob = np.sum(holder)
    cp_memoize[(name1, name2)] = np.exp(log_cnd_prob)
    return cp_memoize


def _probSamePerson(name1, name2, pop_size, edit_count, total_edits, smoothing,
                    ngram_len, ngram_count, memoize, cp_memoize, psp_memoize):
    # computes the probability that the two names belong to the same person.
    if not memoize[name1]:
        memoize = _probName(name1, ngram_len, ngram_count, smoothing, memoize)
    if not memoize[name2]:
        memoize = _probName(name2, ngram_len, ngram_count, smoothing, memoize)
    if not cp_memoize[(name1, name2)]:
        cp_memoize = _condProbName(name1, name2, edit_count, total_edits, smoothing, cp_memoize)
    p1 = memoize[name1]
    p2 = memoize[name2]
    p2given1 = cp_memoize[(name1, name2)]
    if ((pop_size - 1.0) * p1 * p2 + p1 * p2given1):
        psp_memoize[(name1, name2)] = (p1 * p2given1) / ((pop_size - 1.0) * p1 * p2 + p1 * p2given1)
    else:
        psp_memoize[(name1, name2)] = 0.0
    return [psp_memoize, cp_memoize, memoize]




import logging
from copy import deepcopy

from namematch.utils.utils import *

logger = logging.getLogger()


class DataFile():
    '''Parent class for NewDataFile and ExistingDateFile, which house details about the
    data files input for matching.'''

    def __init__(self, validated_data_file_dict):

        for attribute, attribute_value in validated_data_file_dict.items():
            setattr(self, attribute, attribute_value)

    @classmethod
    def load(cls, data_file_dict):
        '''Load a DataFile instance (either a NewDataFile or an ExistingDataFile).

        Args:
            data_file_dict (dict): dictionary-version of a DataFile object

        Returns:
            instance of the DataFile class
        '''

        return cls(data_file_dict)

    def validate_existance(self):
        '''Validate that an input file path exists.'''

        if not os.path.exists(self.filepath):
            logger.error(f"The {self.nickname} file does not exists: "
                         f"{self.filepath}")
            raise

    def validate_record_id_col(self):
        '''Validate that the record_id column exists and meets uniqueness criteria.'''

        df = pd.read_csv(self.filepath, nrows=1)
        if self.record_id_col not in df.columns.tolist():
            logger.error(f"The record_id_col specified for the {self.nickname} "
                         f"file does not exist: {self.record_id_col}")
            raise

        df = pd.read_csv(self.filepath, usecols=[self.record_id_col])
        if df[self.record_id_col].isnull().sum() > 0:
            logger.error(f"The record_id_col specified for the {self.nickname} "
                         f"file contains nulls: {self.record_id_col}.")
            raise
        if len(df) != df[self.record_id_col].nunique():
            logger.warning(f"The record_id_col specified for the {self.nickname} "
                           f"file is not unique: {self.record_id_col}.")
        if (self.file_type == 'existing') and self.use_record_id_as_is:
            split_recids = df[self.record_id_col].str.split("__", n=1, expand=True)
            if (split_recids.shape[1] == 1) or split_recids[1].isnull().any():
                logger.error(f"The use_record_id_as_is parameter is set to True for the {self.nickname} file, "
                             f"but at least one of the underlying record_id values contains no prefix "
                             f"(e.g. 'arrests__'). Consider changing use_record_id_as_is to False.")
                raise
            prefixes = split_recids[0].unique().tolist()
            logger.info(f"The use_record_id_as_is parameter is set to True for the {self.nickname} file. "
                        f"The ids are prefixed by the following nicknames: {', '.join(prefixes)}.")


    def copy(self):
        '''Create a deep copy of a DataFile object.'''
        return deepcopy(self)


class NewDataFile(DataFile):

    @classmethod
    def build(cls, nickname, info):
        '''Create a NewDataFile instance.

        Args:
            nickname (str): the data file's nickname
            info (dict): info about a data file definition from user-input config

        Returns:
           :mod:`namematch.data_structures.data_file.NewDataFile`: instance of the NewDataFile class
        '''

        params_required = ['filepath', 'record_id_col']
        for param_required in params_required:
            if param_required not in info:
                logger.error(f"Error in data_file definition. Please ensure "
                             f"{param_required} is defined.")
                raise ValueError

        data_file_dict = {}
        data_file_dict['nickname'] = nickname
        data_file_dict['filepath'] = info['filepath']
        data_file_dict['record_id_col'] = info['record_id_col']
        data_file_dict['delim'] = info.get('delim', ',')
        data_file_dict['file_type'] = 'new'
        data_file_dict['use_record_id_as_is'] = False
        data_file_dict['cluster_type'] = info.get('cluster_type', 'cluster')
        data_file_dict['output_file_stem'] = info.get('output_file_stem', nickname)

        return cls(data_file_dict)


class ExistingDataFile(DataFile):

    @classmethod
    def build(cls, nickname, info):
        '''Create a ExistingDataFile instance.

        Args:
            nickname (str): the data file's nickname
            info (dict): info about a data file definition from user-input config

        Returns:
            :mod:`namematch.data_structures.data_file.ExistingDataFile`: instance of the ExistingDataFile class
        '''

        params_required = ['filepath', 'record_id_col']
        for param_required in params_required:
            if param_required not in info:
                logger.error('Error in existing_data_file definition. Please ensure %s is defined.' % param_required)
                raise ValueError

        data_file_dict = {}
        data_file_dict['nickname'] = nickname
        data_file_dict['filepath'] = info['filepath']
        data_file_dict['record_id_col'] = info['record_id_col']
        data_file_dict['delim'] = info.get('delim', ',')
        data_file_dict['file_type'] = 'existing'
        data_file_dict['use_record_id_as_is'] = info.get('use_record_id_as_is', False)

        return cls(data_file_dict)


class DataFileList():
    '''Class that houses a list of DataFile objects (either NewDataFiles or ExistingDateFiles.'''

    def __init__(self, data_files_dict):

        self.data_files = data_files_dict['data_files']
        self.existing_data_files = data_files_dict['existing_data_files']
        self.validate_names()


    @classmethod
    def build(cls, data_files_dict, existing_data_files_dict):
        '''Create a DataFileList instance.

        Args:
            data_files_dict (dict): dictionary with "new data file" info from user-input config
            existing_data_files_dict (dict): dictionary with "existing data file" info from user-input config

        Returns:
            :mod:`namematch.data_structures.data_file.DataFileList`: instance of the DataFileList class
        '''

        validated_data_files_dict = {
            'data_files' : [],
            'existing_data_files' : []
        }
        for data_file_name, data_file_info in data_files_dict.items():
            validated_data_file = NewDataFile.build(data_file_name, data_file_info)
            validated_data_files_dict['data_files'].append(validated_data_file)
        for data_file_name, data_file_info in existing_data_files_dict.items():
            validated_data_file = ExistingDataFile.build(data_file_name, data_file_info)
            validated_data_files_dict['existing_data_files'].append(validated_data_file)

        return cls(validated_data_files_dict)


    @classmethod
    def load(cls, data_files_list_dict):
        '''Load a DataFileList instance.

        Args:
            data_files_list_dict (dict): dictionary-version of a DataFileList object

        Returns:
            :mod:`namematch.data_structures.data_file.DataFileList`: instance of the DataFileList class
        '''

        data_files = [
                DataFile.load(data_file_dict)
                for data_file_dict in data_files_list_dict['data_files']]
        existing_data_files = [
                DataFile.load(data_file_dict)
                for data_file_dict in data_files_list_dict['existing_data_files']]

        data_files_dict = {
            'data_files' : data_files,
            'existing_data_files' : existing_data_files
        }

        # NOTE: no need for validation because simple loading
        # an object that was validated when initialized

        return cls(data_files_dict)


    def get_all_nicknames(self):
        '''Return a list of all of the DataFile nicknames in the DataFileList.

        Return:
            list of strings
        '''

        data_file_name_list = [df.nickname for df in self.data_files]
        existing_data_file_name_list = [df.nickname for df in self.existing_data_files]
        if len(existing_data_file_name_list) > 0:
            data_file_name_list.extend(existing_data_file_name_list)

        return data_file_name_list


    def validate(self):
        '''Validate the DataFileList by validating the list overall and then validating
        each individual DataFile.'''

        self.validate_names()
        for data_file in (self.data_files + self.existing_data_files):
            data_file.validate_existance()
            data_file.validate_record_id_col()


    def validate_names(self):
        '''Validate that the DataFiles in the DataFileList all have unique nicknames
        and that the output file stems have unique cluster types.'''

        # check for unique nicknames
        data_file_name_list = self.get_all_nicknames()
        if len(data_file_name_list) != len(set(data_file_name_list)):
            logger.error('Input data files must have unique nicknames.')
            raise ValueError

        # check for unique cluster_types per output_file_stem
        cluster_types = [df.cluster_type for df in self.data_files]
        output_stems = [df.output_file_stem for df in self.data_files]
        name_df = pd.DataFrame(data={
                'cluster_types':cluster_types,
                'output_stems':output_stems})
        cluster_types_per_file = name_df.groupby('output_stems').cluster_types.nunique()
        too_many = cluster_types_per_file[cluster_types_per_file > 1].index.tolist()
        if len(too_many) > 0:
            logger.error(f"Within an output_file_stem, the cluster_types must be "
                         f"unique. Revisit the config specification for the "
                         f"following output_file_stem(s): {data_file.output_file_stem}")
            raise ValueError()


    def write(self, output_file):
        '''Write the DataFileList to a yaml file.

        Args:
            output_file (str): path to write data file list dictionary
        '''

        dump_yaml(to_dict(self), output_file)


    def copy(self):
        '''Create a deep copy of a DataFileList object.'''

        return deepcopy(self)


    def get_all_data_files(self):
        '''Retrieve list of all DataFile objects, regardless of New or Existing.

        Return:
            list of DataFile objects
        '''

        all_data_files = []
        for file_type, data_file_list in self.__dict__.items():
            for data_file in data_file_list:
                all_data_files.append(data_file.copy())

        return all_data_files

